/**
 * \file log.cpp
 * \brief The log interface.
 */
#include "log.h"

#include <cstring>
#include <sstream>

#include <boost/thread/lock_guard.hpp>

using namespace std;

ostream &Log(string module, string level)
{
	// We take for granted it's created if it doesn't exist
	assert(Interfaces::LogManager::getSingletonPtr());

	return Interfaces::LogManager::getSingleton().getStream(module, level);
}

namespace Interfaces {

	LogManager::LogManager()
	{
		mDefaultConfig = DEFAULT_OUTPUT;

		mLastFinished = true;

		ostringstream tmp;
		tmp << endl;
		mNewline = tmp.str();
	}

	LogManager::~LogManager()
	{
		closeFile();
	}

	// LogManager - output part
	bool LogManager::send(LogId id, string message, string file,
			unsigned int line, tm *timePtr)
	{
		ostringstream header;

		// The header is Module: Level [file:line] (date time)
		header << id.first << ": " << id.second;

		if(!file.empty()) {
			header << " [" << file;
			if(line != 0) {
				header << ":" << line;
			}
			header << "]";
		}

		if(!timePtr) {
			time_t rawtime;
			time(&rawtime);
			timePtr = localtime(&rawtime);
		}

		char timeString[DATETIME_MAX];
		strftime(timeString, DATETIME_MAX, "%x, %X", timePtr);
		header <<  " (" << timeString << "): ";

		if(mLastFinished || mLastId != id) {
			message.insert(0, header.str());
		}

		/*
		 * We add a tab to each line.
		 *
		 * However we check that it isn't the last occurrence, because we
		 * haven't always full lines and if it is the last one we just have to
		 * store it.
		 */
		size_t found = 0;
		while((found = message.find(mNewline, found)) != string::npos
				&& found != (message.length() - mNewline.length())) {
			found += mNewline.length();
			message.insert(found, "\t");
		}

		/*
		 * We try not to have strange outputs, however this could not work in
		 * case of streams because output is split in parts.
		 *
		 * We have to keep it until the end of the function so we don't create
		 * a block in which we want to have the mutex locked.
		 */
		boost::lock_guard<boost::recursive_mutex> guard(mOutputMutex);

		/*
		 * We consider the log message ended only if the newline was the last
		 * character of it.
		 */
		mLastFinished = found == (message.length() - mNewline.length());
		mLastId = id;

		// By default assume everything succeeded
		bool ret = true;
		unsigned char config = getConfig(id);

		if((config & OT_STDOUT) && cout.good()) {
			cout.write(message.c_str(), message.length());
			ret &= cout.good();
		}

		if((config & OT_STDERR) && cerr.good()) {
			cerr.write(message.c_str(), message.length());
			ret &= cerr.good();
		}

		if((config & OT_FILE) && mLogFile.is_open() && mLogFile.good()) {
			mLogFile.write(message.c_str(), message.length());
			ret &= mLogFile.good();
		}

		if(config & OT_CONSOLE) {
			mConsoleLog.append(message.c_str(), message.length());
		}

		return ret;
	}

	void LogManager::flushConsole(string &str)
	{
		// We modify data of the string, so stop the send message, for now
		boost::lock_guard<boost::recursive_mutex> guard(mOutputMutex);

		str.erase();
		mConsoleLog.swap(str);
	}

	// LogManager - file managing
	bool LogManager::openFile(string filename)
	{
		if(mLogFile.is_open()) {
			closeFile();
		}

		mLogFile.open(filename.c_str());
		return mLogFile.good();
	}

	void LogManager::closeFile()
	{
		if(mLogFile.is_open()) {
			mLogFile.close();
		}
	}

	// LogManager - managing logs part
	ostream &LogManager::getStream(string module, string level)
	{
		LogId id(module, level);

		if(mStreams.find(id) == mStreams.end()) {
			/*
			 * As suggested, use a guard instead of simply locking the mutex.
			 * The guard is automatically deleted (and therefore the mutex
			 * unlocked) at the end of the if block.
			 */
			boost::lock_guard<boost::mutex> guard(mStreamsMutex);

			/*
			 * Before locking the mutex the streambuf could have been created.
			 * Checking for the existence again saves for sure from memory leaks
			 * and segmentation faults, because it can't be created while the
			 * mutex is locked.
			 */
			if(mStreams.find(id) == mStreams.end()) {
				mStreams[id] = StreamPtr(new ostream(
						LogStreamBuf::getStreambuf(id)));
			}
		}

		return *mStreams[id].get();
	}

	// LogManager - Configuration part
	unsigned char LogManager::getConfig(LogId id)
	{
		// Priority: log > module > level > default
		unsigned char ret = mDefaultConfig;
		getMapConfig(mLevelConfig, id.second, ret);
		getMapConfig(mModuleConfig, id.first, ret);
		getMapConfig(mLogConfig, id, ret);

		return ret;
	}

	void LogManager::setGenericConfig(unsigned char &target, OutputType type,
			bool enable)
	{
		if(enable) {
			// Type has the correct bit high, so it will highed on mConfig
			target |= type;
		} else {
			/*
			 * max XOR destination bit will have all bits high except the
			 * destination bit.
			 * Therefore using AND will change only the bit we want to be low.
			 */
			target &= numeric_limits<unsigned char>::max() ^ type;
		}
	}

	void LogManager::setDefaultConfig(OutputType type, bool enable)
	{
		setGenericConfig(mDefaultConfig, type, enable);
	}

	void LogManager::setModuleConfig(std::string module, OutputType type,
			bool enable)
	{
		setGenericConfig(mModuleConfig[module], type, enable);
	}

	void LogManager::setLevelConfig(std::string level, OutputType type,
			bool enable)
	{
		setGenericConfig(mLevelConfig[level], type, enable);
	}

	void LogManager::setLogConfig(LogId id, OutputType type, bool enable)
	{
		setGenericConfig(mLogConfig[id], type, enable);
	}

	// Streambuf - Getting instances part
	std::map<LogId, LogStreamBuf::Ptr> LogStreamBuf::msStreambufs;
	boost::mutex LogStreamBuf::msMutex;

	LogStreamBuf *LogStreamBuf::getStreambuf(LogId &id)
	{
		if(msStreambufs.find(id) == msStreambufs.end()) {
			// As LogManager::getStream
			boost::lock_guard<boost::mutex> guard(msMutex);

			if(msStreambufs.find(id) == msStreambufs.end()) {
				msStreambufs[id] = Ptr(new LogStreamBuf(id));
			}
		}

		return msStreambufs[id].get();
	}

	LogStreamBuf::LogStreamBuf(LogId &id, size_t bufferSize) {
		mId = id;

		mBuffer = new char[bufferSize];
		memset(mBuffer, 0, bufferSize);
		setp(mBuffer, mBuffer + bufferSize);
	}

	LogStreamBuf::~LogStreamBuf() {
		// Last sync before completely delete all data
		sync();

		delete mBuffer;
		mBuffer = 0;
	}

	// Streambuf - Output part
	streamsize LogStreamBuf::xsputn(const char *s, streamsize n)
	{
		// The argument of memcpy is an unsigned int, so ignore n <= 0
		if(n <= 0) {
			return 0;
		}

		streamsize size, put = 0;

		/*
		 * As described in http://en.cppreference.com/w/cpp/io/basic_streambuf/sputn
		 * if the put area becomes full, this function should achieve the result
		 * of calling overflow.
		 *
		 * We use this while to send the whole string.
		 * We use sync() as warranty that the loop won't be infinite: if sync
		 * successes the area becomes empty, otherwise, if sync fails, the loop
		 * is broken.
		 */
		do {
			// Avoid every possible error with memcpy
			size = min(n, epptr() - pptr());

			// Use memcpy instead of strcpy to allow binary transfers
			memcpy(pptr(), s + put, size);
			pbump(size);

			put += size;
			n -= size;
		} while(n > 0 && !sync());

		return put;
	}

	LogStreamBuf::int_type LogStreamBuf::overflow(int_type c)
	{
		// Should never happen
		if(pbase() > epptr()) {
			return EOF;
		}

		*pptr() = c;
		pbump(1);

		// Sync when we are out of space in the buffer
		if(pptr() >= epptr()) {
			sync();
		}

		return c;
	}

	int LogStreamBuf::sync()
	{
		ptrdiff_t n = pptr() - pbase();

		if(n > 0) {

			string tmp(pbase(), n);
			// Rewind the pointer in any case
			pbump(-n);

			return LogManager::getSingleton().send(mId, tmp) ? 0 : -1;
		} else {
			return 0;
		}
	}

} /* namespace Interfaces */
