/**
 * \file log.h
 * \brief The log interface.
 *
 * This is an interface to use logs with a C++ stream.
 * The Log function is called to get streams.
 */

#ifndef INTERFACES_LOG_H_
#define INTERFACES_LOG_H_

#include <fstream>
#include <string>
#include <ostream>
#include <map>
#include <limits>

#include <cstdio>
#include <ctime>

#include <boost/shared_ptr.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/recursive_mutex.hpp>

#include "singleton.h"

/**
 * \brief Get a log stream
 *
 * This function returns a log stream.
 * It is a function but it is called with our class policy because it acts as
 * an apparent class, as it returns one.
 *
 * \param module: The module which sends the log
 * \param level: The log level (e. g. information, warning...)
 */
std::ostream &Log(std::string module, std::string level);

namespace Interfaces {

/**
 * \brief The log id type. The pair is of module-level.
 */
typedef std::pair<std::string, std::string> LogId;

/**
 * \brief This class manages all the logs streams.
 */
class LogManager : public Singleton<LogManager> {
public:
	/**
	 * \brief The smart pointer to the ostream.
	 *
	 * We use Boost Smart Pointers because we need to store ostreams, but they
	 * are not compatible with the maps containers, therefore we have to use
	 * pointers.
	 *
	 * However we want to be sure that these instances will be deleted so we
	 * decided to use smart pointers.
	 */
	typedef boost::shared_ptr<std::ostream> StreamPtr;

	/**
	 * \brief Possible outputs for logs.
	 *
	 * Console is for the in game console.
	 */
	enum OutputType {
		OT_STDOUT = 1,
		OT_STDERR = 1 << 1,
		OT_FILE = 1 << 2,
		OT_CONSOLE = 1 << 3
	};

	/**
	 * \brief The max length of the date and time.
	 *
	 * Logs print date and time, too, and they do using strftime, which needs a
	 * C string, so we must define the max length of it.
	 */
	static const size_t DATETIME_MAX = 30;

	/**
	 * \brief The default output configuration.
	 */
	static const char DEFAULT_OUTPUT = OT_STDOUT | OT_CONSOLE;

	~LogManager();

	/**
	 * \brief Send a log message.
	 * \param id: The log id
	 * \param message: The message to log
	 * \param file: The file that sent the message
	 * \param line: The line of the file. Ignored if file is empty
	 * \param time: The time of the message. If 0 it is automatically detected
	 * \return: Did it succeed?
	 */
	bool send(LogId id, std::string message, std::string file = "", unsigned int
			line = 0, tm *timePtr = 0);

	/**
	 * \brief Get a stream.
	 * \param module: The module to get the stream of
	 * \param level: The level of the module to get the stream of
	 * \return: The ostream of the log
	 */
	std::ostream &getStream(std::string module, std::string level);

	/**
	 * \brief Get the output configuration for a log.
	 * \param id: The log id
	 * \return: The configuration
	 */
	unsigned char getConfig(LogId id);

	/**
	 * \brief Set the default configuration.
	 * \param type: The destination
	 * \param enabled: Is this destination enabled?
	 */
	void setDefaultConfig(OutputType type, bool enable);

	/**
	 * \brief Set the configuration for a module.
	 * \param module: The module
	 * \param type: The destination
	 * \param enabled: Is this destination enabled?
	 */
	void setModuleConfig(std::string module, OutputType type, bool enable);

	/**
	 * \brief Set the configuration for a level.
	 * \param level: The level
	 * \param type: The destination
	 * \param enabled: Is this destination enabled?
	 */
	void setLevelConfig(std::string level, OutputType type, bool enable);

	/**
	 * \brief Set the configuration for a log.
	 * \param id: The log id
	 * \param type: The destination
	 * \param enabled: Is this destination enabled?
	 */
	void setLogConfig(LogId id, OutputType type, bool enable);

	/**
	 * \brief Flush the string for the console log.
	 *
	 * We're using a reference passage to the function not to create a string
	 * and then return it copying the content of the log too many times.
	 *
	 * \param str: The string that will have the output.
	 */
	void flushConsole(std::string &str);

	/**
	 * \brief Open the log file. If another one is opened, it is closed.
	 * \param filename: The file name
	 * \return: The status of the stream
	 */
	bool openFile(std::string filename);

	/**
	 * \brief Closes the log file.
	 */
	void closeFile();

private:
	/**
	 * \brief The constructor.
	 */
	LogManager();

	/**
	 * \brief Get the configuration from a map.
	 *
	 * We follow the don't reinvent the wheel®, but we can't use only a function
	 * because we have different types of maps so we have to use a template.
	 *
	 * \param map: The map which to search in
	 * \param key: The key to find.
	 */
	template<typename K>
	void getMapConfig(std::map<K, unsigned char> &map, K key,
			unsigned char &config)
	{
		if(map.find(key) != map.end()) {
			config = map[key];
		}
	}

	/**
	 * \brief Enable or disable a destination for the logs.
	 * \param type: The output destination to configure
	 * \param enable: Enable or disable this output?
	 */
	void setGenericConfig(unsigned char &target, OutputType type, bool enable);

	/**
	 * \brief The log streams.
	 *
	 * All the logs are stored in this map.
	 */
	std::map<LogId, StreamPtr> mStreams;

	/**
	 * \brief The mStreams mutex.
	 */
	boost::mutex mStreamsMutex;

	/**
	 * \brief The output mutex.
	 *
	 * We prefer using a recursive mutex because it could be that there is an
	 * exception while outputting the log, and its handler could want to send
	 * some log, so we prefer avoiding deadlocks.
	 *
	 * We won't likely have this problem with other mutexes so we declare only
	 * this one as an explicit recursive mutex.
	 */
	boost::recursive_mutex mOutputMutex;

	/**
	 * \brief The output file stream.
	 */
	std::fstream mLogFile;

	/**
	 * \brief The string for the console.
	 */
	std::string mConsoleLog;

	/**
	 * \brief Was last message finished (had it endl at the end?)?
	 */
	bool mLastFinished;

	/**
	 * \brief The id of the last log which sent output.
	 */
	LogId mLastId;

	/**
	 * \brief The endl content.
	 */
	std::string mNewline;

	/**
	 * \brief The default configuration.
	 */
	unsigned char mDefaultConfig;

	/**
	 * \brief The configuration for each module.
	 */
	std::map<std::string, unsigned char> mModuleConfig;

	/**
	 * \brief The configuration for each level.
	 */
	std::map<std::string, unsigned char> mLevelConfig;

	/**
	 * \brief The configuration for each log.
	 */
	std::map<LogId, unsigned char> mLogConfig;

// Allow singleton creation.
friend class AutoSingleton<LogManager>;
};

/**
 * \brief The streambuf for Logs.
 *
 * Standard C++ streams need a "stream buffer", a class which has to handle
 * the input and output buffers.
 *
 * This stream buffer is actually a class that calls other stream buffers, which
 * can be, depending on its configuration, standard error, a file and a string
 * buffer.
 *
 * This class is like a Singleton: it can be instanced only by itself, to make
 * sure that there will be only a strembuf for each pair module-level.
 */
class LogStreamBuf : public std::streambuf {
public:
	/**
	 * \brief The smart pointer to this object.
	 *
	 * We want to store these instances, but it would be difficult with maps.
	 * So we decided to use smart pointers, even though it isn't likely for that
	 * to be deleted, as they are stored in a static map.
	 */
	typedef boost::shared_ptr<LogStreamBuf> Ptr;

	/**
	 * \brief The default size of the buffer.
	 */
	static const std::size_t DEFAULT_SIZE = 1024;

	/**
	 * \brief Get an instance.
	 * \param id: The log id
	 */
	static LogStreamBuf *getStreambuf(LogId &id);

	/**
	 * \brief The destructor.
	 */
	~LogStreamBuf();

private:
	/**
	 * \brief The constructor. Stores information of the streambuf.
	 * \param id: The log id
	 * \param bufferSize: The size of the output buffer.
	 */
	LogStreamBuf(LogId &id, std::size_t bufferSize = DEFAULT_SIZE);

	/**
	 * \brief The existent streams.
	 */
	static std::map<LogId, Ptr> msStreambufs;

	/**
	 * \brief The mutex to lock the creation of streambufs.
	 */
	static boost::mutex msMutex;

	/**
	 * \brief The log id.
	 */
	LogId mId;

	/**
	 * \brief The buffer where data is saved.
	 *
	 * This is required by the standard streambuf.
	 */
	char *mBuffer;

protected:

	/**
	 * \brief Put a sequence of characters.
	 *
	 * The behavior with invalid arguments is undefined, for the standard.
	 * If the string is wrong, this function will copy it nevertheless, maybe
	 * memcpy could fail.
	 * It the length <= 0, then the function will simply return 0.
	 *
	 * \param s: The string to output, not necessarily ending with null char.
	 * \param n: The length of the string. Should be positive.
	 * \return The number of characters that have been written.
	 */
	virtual std::streamsize xsputn(const char *s, std::streamsize n);

	/**
	 * \brief Put a single character on the buffer.
	 * \param c: The character to put.
	 * \return c, or EOF in case of error
	 */
	virtual int_type overflow(int_type c = EOF);

	/**
	 * \brief Synchronize the stream buffer
	 * \return 0 in case of success, -1 otherwise
	 */
	virtual int sync();
};

} /* namespace Interfaces */

#endif /* INTERFACES_LOG_H_ */
