/**
 * \file singleton.h
 * \brief A general way to implement the Singleton pattern.
 */

#ifndef INTERFACES_SINGLETON_H_
#define INTERFACES_SINGLETON_H_

#include <cassert>

namespace Interfaces {

	/**
	 * \brief The class to allow the Singleton auto creation.
	 *
	 * The Singleton template needs to have a fall back action in case the
	 * Singleton doesn't exist, and this is the aim of this class.
	 * This class creates the object if it doesn't exist.
	 *
	 * Please remember to create a friendship between T and AutoSingleton<T>!
	 *
	 * \see Singleton
	 */
	template<typename T>
	class AutoSingleton {
	public:
		/**
		 * \brief Create a T object and return it.
		 */
		inline static T *create()
		{
			return new T();
		}
	};

	/**
	 * \brief The class to return a null object if the Singleton doesn't exist.
	 * \see Singleton
	 * \see AutoSingleton
	 */
	template<typename T>
	class NoAutoSingleton {
	public:
		/**
		 * \brief Return the null pointer.
		 */
		inline static T *create()
		{
			return 0;
		}
	};

	/**
	 * \brief The interface to create singletons.
	 *
	 * T is the class that must be a singleton.
	 *
	 * The creation can be done automatically by this class, however it can be
	 * disabled because some classes are just interfaces but their
	 * implementation must be a singleton.
	 * Sadly the feature can be implemented only using classes to create the
	 * object. Using an if would make the compiler complain.
	 *
	 * Only one template must exist per type, otherwise the system doesn't work.
	 */
	template<typename T, template <typename> class C = AutoSingleton>
	class Singleton {
	protected:
		/**
		 * \brief The constructor to not allow this class to be instanced.
		 */
		Singleton() { }

		/**
		 * \brief The singleton handler.
		 */
		static T *msSingleton;

		/**
		 * \brief Set the current object as Singleton.
		 *
		 * We just cast the this, we don't do further tests because we take for
		 * granted that it's called by a child class, since it's protected, and
		 * that this child class parent is a child of Singleton<T>.
		 */
		void setSingleton()
		{
			assert(!msSingleton);
			msSingleton = static_cast<T *>(this);
		}

	public:
		/**
		 * \brief Get the singleton pointer.
		 *
		 * Warning: the singleton pointer could not be valid, so be cautious.
		 */
		static T *getSingletonPtr()
		{
			if(!msSingleton) {
				msSingleton = C<T>::create();
			}

			return msSingleton;
		}

		/**
		 * \brief Get the singleton instance.
		 */
		static T &getSingleton()
		{
			// We don't want exceptions there!
			assert(getSingletonPtr());
			return *getSingletonPtr();
		}
	};

	template<typename T, template <typename> class C>
		T *Interfaces::Singleton<T, C>::msSingleton = 0;
}

#endif /* INTERFACES_SINGLETON_H_ */
